/*
 * Copyright (c) 2016-2017 Intel Corporation. All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef DATA_CONTAINER_H
#define DATA_CONTAINER_H

#include <stdexcept>
#include <sys/time.h>
#include <typeinfo>

#include "dataHolder.hpp"

const std::string KEY("key");
const std::string VALUE("value");
const std::string UNITS("units");
const std::string TIMESTAMP("timestamp");

// Exceptions
class unableToFindKey : public std::runtime_error {
    public:
        unableToFindKey(const std::string &k) : std::runtime_error(
            std::string("Unable to find key in container (") + k + std::string(")")) {}
};

class unableToFindData : public std::runtime_error {
    public:
        unableToFindData(const std::string &d) : std::runtime_error(
            std::string("Unable to find data in container (") + d + std::string(")")) {}
};

class reachedEndOfContainer : public std::runtime_error {
    public:
        reachedEndOfContainer(void) : std::runtime_error(
            std::string("Tried to access the container with an out-of-range iterator.")) {}
};

class dataContainer {
    public:
        typedef std::vector< dataHolderMap >::iterator iterator;

        dataContainer() {};

        template<typename T> inline dataContainer::iterator put(const std::string& k,
                const T &v, const std::string &u) {
            dataHolderMap map;
            map[KEY] = dataHolder(k);
            map[VALUE] = dataHolder(v);
            container.push_back(map);

            dataContainer::iterator last = container.end()-1;
            setMetadata(last, UNITS, u);
            setMetadata(last, TIMESTAMP, getTimestamp());
            return last;
        }

        template<typename T> inline void setMetadata(const dataContainer::iterator& it,
                const std::string& name, const T &value) {
            verifyIterator(it);
            (*it)[name] = dataHolder(value);
        }

        template<typename T> inline void setMetadata(const std::string& key,
                const std::string& name, const T &value) {
            setMetadata(getIteratorFromStringKeyAndCheck(key), name, value);
        }

        template<typename T> inline T getMetadata(const dataContainer::iterator& it,
                const std::string& name) {
            verifyIterator(it);
            if (it->find(name) == it->end())
                throw unableToFindData(name);
            return it->at(name).getValue<T>();
        }

        template<typename T> inline T getMetadata(const std::string& key, const std::string& name) {
            return getMetadata<T>(getIteratorFromStringKeyAndCheck(key), name);
        }

        template<typename T> inline T getValue(const std::string& key) {
            return getValue<T>(getIteratorFromStringKeyAndCheck(key));
        }

        template<typename T> inline T getValue(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(VALUE).getValue<T>();
        }

        inline void erase(const dataContainer::iterator& it) {
            verifyIterator(it);
            container.erase(it);
        }

        inline void erase(const std::string& key) {
            this->erase(getIteratorFromStringKeyAndCheck(key));
        }

        inline void concat(dataContainer &dc) {
            (this->container).insert(this->container.begin(), dc.begin(), dc.end());
        }

        inline void clear() {
            (this->container).clear();
        }

        inline bool containsKey(const std::string& key) {
            return !(getIteratorFromStringKey(key) == container.end());
        };

        inline size_t getDataSize(const std::string& key) {
            dataContainer::iterator it = getIteratorFromStringKeyAndCheck(key);
            return it->at(VALUE).size();
        };

        inline size_t count() {
            return container.size();
        };

        inline dataContainer::iterator begin() {
            return container.begin();
        };

        inline dataContainer::iterator end() {
            return container.end();
        };

        inline const std::string getKey(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(KEY).getValue<std::string>();
        };

        inline const std::string getUnits(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(UNITS).getValue<std::string>();
        };

        inline const std::string getUnits(const std::string& key) {
            return getUnits(getIteratorFromStringKeyAndCheck(key));
        };

        inline const std::string getDataTypeName(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(VALUE).getDataTypeName();
        };

        inline const std::string getDataTypeName(const std::string& key) {
            return getDataTypeName(getIteratorFromStringKeyAndCheck(key));
        };

        template<typename T> inline bool matchType(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(VALUE).is<T>();
        };

        template<typename T> inline bool matchType(const std::string& key) {
            return matchType<T>(getIteratorFromStringKey(key));
        };

        inline void* getRawDataPtr(const dataContainer::iterator& it) const {
            verifyIterator(it);
            return it->at(VALUE).ptr<void>();
        };

    protected:
        std::vector< dataHolderMap > container;

    private:
        inline dataContainer::iterator getIteratorFromStringKey(const std::string& key) {
            dataContainer::iterator it = container.begin();
            for (; it != container.end(); ++it) {
                if (it->find(KEY) != it->end() &&
                    0 == it->at(KEY).getValue<std::string>().compare(key))
                    break;
            }
            return it;
        }

        inline dataContainer::iterator getIteratorFromStringKeyAndCheck(const std::string& key) {
            dataContainer::iterator it = getIteratorFromStringKey(key);
            if (it == container.end())
                throw unableToFindKey(key);
            return it;
        };

        struct timeval getTimestamp(void) {
            struct timeval current_time;
            gettimeofday(&current_time, NULL);
            return current_time;
        }

        inline void verifyIterator(const dataContainer::iterator& it) const {
            if (it == container.end())
                throw reachedEndOfContainer();
        }

};

typedef std::map<std::string, dataContainer> dataContainerMap;

#endif
